using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Agreement;


public class DHExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("==================================");
        Console.WriteLine("Start Diffie-Hellman Examples.Main");
        Console.WriteLine("==================================");
        Console.WriteLine("==================================");


        {
            //Example 19 – Generating DH Finite Field Parameters
            Console.WriteLine();
            Console.WriteLine("*** Example 19 – Generating DH Finite Field Parameters ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            DHParameters dhParams = generateDHParameters(256);
            Console.WriteLine(string.Format("DH Parameters\tP:[{0}]\tG:[{1}]\tQ:[{2}]",
                dhParams.P.ToString(16),
                dhParams.G.ToString(16),
                dhParams.Q.ToString(16)
                ));
        }
        
        
        {
            //Example 20 – Diffie-Hellman Key Pair Generation
            Console.WriteLine();
            Console.WriteLine("*** Example 20 – Diffie-Hellman Key Pair Generation ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
            DHParameters dhParams = DHStandardGroups.rfc7919_ffdhe3072;           

            DHKeyGenerationParameters dhKeyGenParams = 
               new DHKeyGenerationParameters(ExValues.cSharpFixedRandom, dhParams);
            DHKeyPairGenerator dhKeyPairGen = new DHKeyPairGenerator();
            dhKeyPairGen.Init(dhKeyGenParams);
            
       		AsymmetricCipherKeyPair dhKeyPair = dhKeyPairGen.GenerateKeyPair();
			DHPublicKeyParameters dhPublicKey   = (DHPublicKeyParameters)dhKeyPair.Public;
			DHPrivateKeyParameters dhPrivateKey = (DHPrivateKeyParameters)dhKeyPair.Private;
            Console.WriteLine(string.Format("DH Key Pair Generation\tPublic:[{0}]\tLen:[{1}]\tPrivate:[{2}]\tLen:[{3}]",
                dhPublicKey.Y.ToString(16), dhPublicKey.Y.ToString(16).Length,
                dhPrivateKey.X.ToString(16), dhPrivateKey.X.ToString(16).Length
                ));

       }        
       
       
       {
            //Example 20X – Diffie-Hellman Key Pair Generation
            Console.WriteLine();
            Console.WriteLine("*** Example 20X – Diffie-Hellman Key Pair Generation ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
            DHParameters dhParams = DHStandardGroups.rfc7919_ffdhe3072;           

       		AsymmetricCipherKeyPair dhKeyPair = generateDHKeyPair(dhParams);
			DHPublicKeyParameters dhPublicKey   = (DHPublicKeyParameters)dhKeyPair.Public;
			DHPrivateKeyParameters dhPrivateKey = (DHPrivateKeyParameters)dhKeyPair.Private;
            Console.WriteLine(string.Format("DH Key Pair Generation\tPublic:[{0}]\tLen:[{1}]\tPrivate:[{2}]\tLen:[{3}]",
                dhPublicKey.Y.ToString(16), dhPublicKey.Y.ToString(16).Length,
                dhPrivateKey.X.ToString(16), dhPrivateKey.X.ToString(16).Length
                ));
           
       }

       {
            //Example 21 – Diffie-Hellman Basic Agreement
            Console.WriteLine();
            Console.WriteLine("*** Example 21 – Diffie-Hellman Basic Agreement ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
       
            DHParameters dhParams = DHStandardGroups.rfc7919_ffdhe3072;           
       		AsymmetricCipherKeyPair dhKeyPairPartyA = generateDHKeyPair(dhParams);
       		AsymmetricCipherKeyPair dhKeyPairPartyB = generateDHKeyPair(dhParams);

            BigInteger secretPartyA = partyABasicAgreement(
                (DHPrivateKeyParameters)dhKeyPairPartyA.Private,
                (DHPublicKeyParameters)dhKeyPairPartyB.Public);
    
            BigInteger secretPartyB = partyBBasicAgreement(
                (DHPrivateKeyParameters)dhKeyPairPartyB.Private,
                (DHPublicKeyParameters)dhKeyPairPartyA.Public);

            bool result = secretPartyA.Equals(secretPartyB);
            
            Console.WriteLine(string.Format("DH Secret Agreement A\t:[{0}]\tLen:[{1}]",
                secretPartyA.ToString(16), secretPartyA.ToString(16).Length
                ));
            Console.WriteLine(string.Format("DH Secret Agreement B\t:[{0}]\tLen:[{1}]",
                secretPartyB.ToString(16), secretPartyB.ToString(16).Length
                ));
            Console.WriteLine(string.Format("DH Secret A Agrees With Secret B\t:[{0}]",result));
       
        }
        
       {
            //Example 22 – Matsumoto-Takashima-Imai MTI/A0 Agreement
            Console.WriteLine();
            Console.WriteLine("*** Example 22 – Matsumoto-Takashima-Imai MTI/A0 Agreement ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
       
            DHParameters dhParams = DHStandardGroups.rfc7919_ffdhe3072;           
            BigInteger[] secrets = MTIA0Agreement(dhParams);
            
            bool result = secrets[0].Equals(secrets[1]);
            
            Console.WriteLine(string.Format("DH Secret Agreement A\t:[{0}]\tLen:[{1}]",
                secrets[0].ToString(16), secrets[0].ToString(16).Length
                ));
            Console.WriteLine(string.Format("DH Secret Agreement B\t:[{0}]\tLen:[{1}]",
                secrets[1].ToString(16), secrets[1].ToString(16).Length
                ));
            Console.WriteLine(string.Format("DH Secret A Agrees With Secret B\t:[{0}]",result));
            
       }        
        Console.WriteLine("==================================");
        Console.WriteLine("End   Diffie-Hellman Examples.Main");
        Console.WriteLine("==================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    

    /// <summary>
    /// Example 19 – Generating DH Finite Field Parameters
    /// </summary>        
    public static DHParameters generateDHParameters(int size)
    {
        DHParametersGenerator pGen = new DHParametersGenerator();
        pGen.Init(size, 10, ExValues.cSharpFixedRandom);  
        return pGen.GenerateParameters();
    }

      
    /// <summary>
    /// Example 20 – Diffie-Hellman Key Pair Generation
    /// </summary>        
    public static DHKeyParameters[] generateDHKeyPairArray(DHParameters dhParams)
    {
        DHKeyGenerationParameters dhKeyGenParams = 
           new DHKeyGenerationParameters(ExValues.cSharpFixedRandom, dhParams);
        DHKeyPairGenerator dhKeyPairGen = new DHKeyPairGenerator();
        dhKeyPairGen.Init(dhKeyGenParams);
            
        AsymmetricCipherKeyPair dhKeyPair = dhKeyPairGen.GenerateKeyPair();
        DHPublicKeyParameters dhPublicKey   = (DHPublicKeyParameters)dhKeyPair.Public;
        DHPrivateKeyParameters dhPrivateKey = (DHPrivateKeyParameters)dhKeyPair.Private;
        DHKeyParameters[] result = {dhPublicKey,dhPrivateKey};
        return result;
    }  

    /// <summary>
    /// Example 20X – Diffie-Hellman Key Pair Generation
    /// </summary>        
    public static AsymmetricCipherKeyPair generateDHKeyPair(DHParameters dhParams)
    {
        DHKeyGenerationParameters dhKeyGenParams = 
           new DHKeyGenerationParameters(ExValues.cSharpFixedRandom, dhParams);
        DHKeyPairGenerator dhKeyPairGen = new DHKeyPairGenerator();
        dhKeyPairGen.Init(dhKeyGenParams);
            
        AsymmetricCipherKeyPair dhKeyPair = dhKeyPairGen.GenerateKeyPair();
        return dhKeyPair;
    }  

    /// <summary>
    /// Example 21 – Basic DH Key Agreement
    /// </summary>        
    public static BigInteger partyABasicAgreement(DHPrivateKeyParameters privateKeyPartyA,
                                                  DHPublicKeyParameters publicKeyPartyB)
    {
        DHBasicAgreement keyAgreement = new DHBasicAgreement();
        keyAgreement.Init(privateKeyPartyA);
        BigInteger secret = keyAgreement.CalculateAgreement(publicKeyPartyB);
        
        return secret;
    }
    public static BigInteger partyBBasicAgreement(DHPrivateKeyParameters privateKeyPartyB,
                                                  DHPublicKeyParameters publicKeyPartyA)
    {
        DHBasicAgreement keyAgreementParty = new DHBasicAgreement();
        keyAgreementParty.Init(privateKeyPartyB);
        BigInteger secret = keyAgreementParty.CalculateAgreement(publicKeyPartyA);
        
        return secret;
    }


    /// <summary>
    /// Example 22 – MTI/A0 DH Key Agreement
    /// </summary>            
    public static BigInteger[] MTIA0Agreement(DHParameters dhParams)
    {
        AsymmetricCipherKeyPair dhStaticKeyPairPartyA = generateDHKeyPair(dhParams);
       	AsymmetricCipherKeyPair dhStaticKeyPairPartyB = generateDHKeyPair(dhParams);

        // Party A Generates an ephemeral key pair sending it to Party B
        DHAgreement keyAgreementPartyA = new DHAgreement();
        keyAgreementPartyA.Init((DHPrivateKeyParameters)dhStaticKeyPairPartyA.Private);
        BigInteger dhEphemeralPublicKeyA = keyAgreementPartyA.CalculateMessage();

        // Party B Generates an ephemeral key pair sending it to Party A
        DHAgreement keyAgreementPartyB = new DHAgreement();
        keyAgreementPartyB.Init((DHPrivateKeyParameters)dhStaticKeyPairPartyB.Private);
        BigInteger dhEphemeralPublicKeyB = keyAgreementPartyB.CalculateMessage();
        
        BigInteger secretPartyA = keyAgreementPartyA.CalculateAgreement(
                (DHPublicKeyParameters)dhStaticKeyPairPartyB.Public, dhEphemeralPublicKeyB);
                
        BigInteger secretPartyB = keyAgreementPartyB.CalculateAgreement(
                (DHPublicKeyParameters)dhStaticKeyPairPartyA.Public, dhEphemeralPublicKeyA);
            
        return  new BigInteger[] {secretPartyA, secretPartyB};
    }
 

}
