using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Math.EC;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Agreement;
using Org.BouncyCastle.Asn1.X9;


public class ECExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=====================================");
        Console.WriteLine("Start EC Diffie-Hellman Examples.Main");
        Console.WriteLine("=====================================");
 

        {
            //Example 23 – Generating EC brainpoolP160r1 Directly
            Console.WriteLine();
            Console.WriteLine("*** Example 23 – Generating EC brainpoolP160r1 Directly ***");
            ECDomainParameters ecDomainParams = ecDomainParametersDirect();           
            Console.WriteLine(string.Format("point.X=[{0}]\tpoint.Y=[{1}]",ecDomainParams.G.XCoord.ToBigInteger(),ecDomainParams.G.YCoord.ToBigInteger()));
        }

        {
            //Example 24 – Generating Built-In EC Parameters
            Console.WriteLine();
            Console.WriteLine("*** Example 24 – Generating EC brainpoolP160r1 Built-in ***");
            ECDomainParameters ecDomainParams = ecBuiltInDomainParameters();           
            Console.WriteLine(string.Format("point.X=[{0}]\tpoint.Y=[{1}]",ecDomainParams.G.XCoord.ToBigInteger(),ecDomainParams.G.YCoord.ToBigInteger()));
        }

        {
            //Example 25 – Generating Binary EC sect283k1 Directly
            Console.WriteLine();
            Console.WriteLine("*** Example 25 – Generating Binary EC sect283k1 Directly ***");
            ECDomainParameters ecDomainParams = ecBinaryDomainParametersDirect();           
            Console.WriteLine(string.Format("point.X=[{0}]\tpoint.Y=[{1}]",ecDomainParams.G.XCoord.ToBigInteger(),ecDomainParams.G.YCoord.ToBigInteger()));
        }

        {
            //Example 26 – Generating Built-In Binary EC Parameters
            Console.WriteLine();
            Console.WriteLine("*** Example 26 – Generating Binary EC sect283k1 Built-in ***");
            ECDomainParameters ecDomainParams = ecBuiltInBinaryDomainParameters();           
            Console.WriteLine(string.Format("point.X=[{0}]\tpoint.Y=[{1}]",ecDomainParams.G.XCoord.ToBigInteger(),ecDomainParams.G.YCoord.ToBigInteger()));
        }
        
        {
            //Example 27 – EC Diffie-Hellman Key Pair Generation
            Console.WriteLine();
            Console.WriteLine("*** Example 27 – EC Diffie-Hellman Key Pair Generation ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
            ECDomainParameters ecDomainParams = ecBuiltInBinaryDomainParameters();           

       		AsymmetricCipherKeyPair ecKeyPair = generateECDHKeyPair(ecDomainParams);
			ECPublicKeyParameters ecPublicKey   = (ECPublicKeyParameters)ecKeyPair.Public;
			ECPrivateKeyParameters ecPrivateKey = (ECPrivateKeyParameters)ecKeyPair.Private;
            Console.WriteLine(string.Format("EC Public Key\tx:[{0}]\n\t\ty:[{1}]",
                ecPublicKey.Q.XCoord.ToBigInteger(), ecPublicKey.Q.YCoord.ToBigInteger()));
            Console.WriteLine(string.Format("EC Private Key\td:[{0}]",ecPrivateKey.D));
        }

       {
            //Example 28 – EC Diffie-Hellman Basic Agreement
            Console.WriteLine();
            Console.WriteLine("*** Example 28 – EC Diffie-Hellman Basic Agreement ***");
            ExValues.cSharpFixedRandom.SetSeed(151);            
            ECDomainParameters ecDomainParams = ecBuiltInBinaryDomainParameters();           
       
       		AsymmetricCipherKeyPair ecKeyPairPartyA = generateECDHKeyPair(ecDomainParams);
       		AsymmetricCipherKeyPair ecKeyPairPartyB = generateECDHKeyPair(ecDomainParams);

            BigInteger secretPartyA = partyABasicAgreement(
                (ECPrivateKeyParameters)ecKeyPairPartyA.Private,
                (ECPublicKeyParameters)ecKeyPairPartyB.Public);
    
            BigInteger secretPartyB = partyBBasicAgreement(
                (ECPrivateKeyParameters)ecKeyPairPartyB.Private,
                (ECPublicKeyParameters)ecKeyPairPartyA.Public);

            bool result = secretPartyA.Equals(secretPartyB);
            
            Console.WriteLine(string.Format("ECDH Secret Agreement A\t:[{0}]",secretPartyA));
            Console.WriteLine(string.Format("ECDH Secret Agreement B\t:[{0}]",secretPartyB));
            Console.WriteLine(string.Format("ECDH Secret A Agrees With Secret B\t:[{0}]",result));
       
        }

        Console.WriteLine("=====================================");
        Console.WriteLine("End   EC Diffie-Hellman Examples.Main");
        Console.WriteLine("=====================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    /// <summary>
    /// Example 23 – Generating EC GF(p) Parameters Directly
    /// </summary>            
    public static ECDomainParameters ecDomainParametersDirect()
    {
        //Generating EC brainpoolP160r1
        BigInteger p = new BigInteger(1,Hex.DecodeStrict("E95E4A5F737059DC60DFC7AD95B3D8139515620F"));
        BigInteger a = new BigInteger(1,Hex.DecodeStrict("340E7BE2A280EB74E2BE61BADA745D97E8F7C300"));
        BigInteger b = new BigInteger(1,Hex.DecodeStrict("1E589A8595423412134FAA2DBDEC95C8D8675E58"));
        BigInteger n = new BigInteger(1,Hex.DecodeStrict("E95E4A5F737059DC60DF5991D45029409E60FC09"));
        BigInteger h = BigInteger.One;
        BigInteger generatorX = new BigInteger(1,Hex.DecodeStrict("BED5AF16EA3F6A4F62938C4631EB5AF7BDBCDBC3"));
        BigInteger generatorY = new BigInteger(1,Hex.DecodeStrict("1667CB477A1A8EC338F94741669C976316DA6321"));

        ECCurve curve = new FpCurve(p, a, b, n, h);
        ECPoint generatorPoint = curve.CreatePoint(generatorX, generatorY);
        return new ECDomainParameters(curve, generatorPoint, n, h);
    }
 
    /// <summary>
    /// Example 24 – Generating Built-In EC GF(p) Parameters
    /// </summary>            
    public static ECDomainParameters ecBuiltInDomainParameters()
    {
        X9ECParameters ecParams = ECNamedCurveTable.GetByName("brainpoolp160r1");
        return new ECDomainParameters(ecParams.Curve, ecParams.G, ecParams.N, ecParams.H, ecParams.GetSeed()); 
    }
    
    /// <summary>
    /// Example 25 – Generating Binary EC Parameters Directly
    /// </summary>            
    public static ECDomainParameters ecBinaryDomainParametersDirect()
    {
        //Generating Binary EC sect283k1
        int m  = 283;
        int k1 = 5;
        int k2 = 7;
        int k3 = 12;
        BigInteger a = BigInteger.Zero;
        BigInteger b = BigInteger.One;
        BigInteger n = new BigInteger("1ffffffffffffffffffffffffffffffffffe9ae2ed07577265dff7f94451e061e163c61",16);
        BigInteger h = BigInteger.Four;
        BigInteger generatorX = new BigInteger("503213f78ca44883f1a3b8162f188e553cd265f23c1567a16876913b0c2ac2458492836",16);
        BigInteger generatorY = new BigInteger("1ccda380f1c9e318d90f95d07e5426fe87e45c0e8184698e45962364e34116177dd2259",16);

        ECCurve curve = new F2mCurve(m, k1, k2, k3, a, b, n, h);
        ECPoint generatorPoint = curve.CreatePoint(generatorX, generatorY);
        
        return new ECDomainParameters(curve, generatorPoint, n, h);
    }
    
     
    /// <summary>
    /// Example 26 – Generating Built-In Binary EC GF(p) Parameters
    /// </summary>            
    public static ECDomainParameters ecBuiltInBinaryDomainParameters()
    {
        X9ECParameters ecParams = ECNamedCurveTable.GetByName("K-283");
        return new ECDomainParameters(ecParams.Curve, ecParams.G, ecParams.N, ecParams.H, ecParams.GetSeed()); 
    }

    /// <summary>
    /// Example 27 – EC Diffie-Hellman Key Pair Generation
    /// </summary>        
    public static AsymmetricCipherKeyPair generateECDHKeyPair(ECDomainParameters ecParams)
    {
        ECKeyGenerationParameters ecKeyGenParams = 
           new ECKeyGenerationParameters(ecParams, ExValues.cSharpFixedRandom);
        ECKeyPairGenerator ecKeyPairGen = new ECKeyPairGenerator();
        ecKeyPairGen.Init(ecKeyGenParams);
            
        AsymmetricCipherKeyPair ecKeyPair = ecKeyPairGen.GenerateKeyPair();
        return ecKeyPair;
    }  

    /// <summary>
    /// Example 28 – Basic EC DH Key Agreement
    /// </summary>        
    public static BigInteger partyABasicAgreement(ECPrivateKeyParameters privateKeyPartyA,
                                                  ECPublicKeyParameters publicKeyPartyB)
    {
        ECDHCBasicAgreement keyAgreement = new ECDHCBasicAgreement();
        keyAgreement.Init(privateKeyPartyA);
        BigInteger secret = keyAgreement.CalculateAgreement(publicKeyPartyB);
        
        return secret;
    }
    public static BigInteger partyBBasicAgreement(ECPrivateKeyParameters privateKeyPartyB,
                                                  ECPublicKeyParameters publicKeyPartyA)
    {
        ECDHCBasicAgreement keyAgreementParty = new ECDHCBasicAgreement();
        keyAgreementParty.Init(privateKeyPartyB);
        BigInteger secret = keyAgreementParty.CalculateAgreement(publicKeyPartyA);
        
        return secret;
    }

}
