using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Utilities.IO;


public class StreamCipherExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Stream Cipher Examples.Main");
        Console.WriteLine("=================================");

        {
            //Example 13 – ChaCha Stream Cipher
            Console.WriteLine();
            Console.WriteLine("*** Example 13 – ChaCha Stream Cipher ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            ParametersWithIV paramsWithIV = keyParameterGenerationWithIV(256, ExValues.sampleShortIVnonce);

            byte[] cipherTextData =  chachaStreamEncrypt(paramsWithIV, 
                ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha Stream")).ToArray());
            Console.WriteLine(string.Format("Encrypt ChaCha Stream\tplainTextData:[{0}]\tLength:[{1}]\tcipherTextData:[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha Stream")).ToArray()),
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha Stream")).ToArray()).Length,
                string.Join(",",cipherTextData),
                cipherTextData.Length));
            byte[] plainTextData = chachaStreamDecrypt(paramsWithIV, cipherTextData);
            Console.WriteLine(string.Format("Decrypt ChaCha Stream\tcipherTextData:[{0}]\tLength:[{1}]\tplainTextData:[{2}]\tLength:[{3}]",
                string.Join(",",cipherTextData),
                cipherTextData.Length.ToString(),
                Encoding.Default.GetString(plainTextData),
                Encoding.Default.GetString(plainTextData).Length
                ));
            
       }        
       
        {
            //Example 14 – H128 Stream Cipher
            Console.WriteLine();
            Console.WriteLine("*** Example 14 – H128 Stream Cipher ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            ParametersWithIV paramsWithIV = keyParameterGenerationWithIV(128, ExValues.sampleIVnonce);

            byte[] cipherTextData =  h128StreamEncrypt(paramsWithIV, 
                ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" H128 Stream")).ToArray());
            Console.WriteLine(string.Format("Encrypt H128 Stream\tplainTextData:[{0}]\tLength:[{1}]\tcipherTextData:[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" H128 Stream")).ToArray()),
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" H128 Stream")).ToArray()).Length,
                string.Join(",",cipherTextData),
                cipherTextData.Length));
            byte[] plainTextData = h128StreamDecrypt(paramsWithIV, cipherTextData);
            Console.WriteLine(string.Format("Decrypt H128 Stream\tcipherTextData:[{0}]\tLength:[{1}]\tplainTextData:[{2}]\tLength:[{3}]",
                string.Join(",",cipherTextData),
                cipherTextData.Length.ToString(),
                Encoding.Default.GetString(plainTextData),
                Encoding.Default.GetString(plainTextData).Length
                ));
                
       }        

       
        Console.WriteLine("=================================");
        Console.WriteLine("End   Stream Cipher Examples.Main");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    /// <summary>
    /// Generate appropriate random key 
    /// </summary>        
    public static KeyParameter keyParameterGeneration(int keySize)
    {
        CipherKeyGenerator keyGen = new CipherKeyGenerator();        
        keyGen.Init(new KeyGenerationParameters(ExValues.cSharpFixedRandom, keySize));
        KeyParameter keyParam = keyGen.GenerateKeyParameter();
        return keyParam;
    }

    /// <summary>
    /// Generate key from a given set of bytes
    /// </summary>        
    public static KeyParameter keyParameterGeneration(byte[] myKey)
    {
        KeyParameter keyParam = new KeyParameter(myKey);        
        return keyParam;
    }
    
    /// <summary>
    /// Key and Initialisation Vector Generation
    /// </summary>        
    public static ParametersWithIV keyParameterGenerationWithIV(int keySize, byte[] IV)
    {
        CipherKeyGenerator keyGen = new CipherKeyGenerator();        
        keyGen.Init(new KeyGenerationParameters(ExValues.cSharpFixedRandom, keySize));
        KeyParameter keyParam = keyGen.GenerateKeyParameter();
        ParametersWithIV keyParamWithIV = new ParametersWithIV(keyParam, IV);
        return keyParamWithIV;
    }


    /// <summary>
    /// Example 13 – ChaCha Stream Cipher Encryption
    /// </summary>    
    public static byte[] chachaStreamEncrypt(ICipherParameters keyParamWithIV, byte[] plainTextData)
    {
        IStreamCipher cipher = new ChaChaEngine();
        cipher.Init(true, keyParamWithIV);

        byte[] cipherTextData = new byte[plainTextData.Length];
        
        // simulate stream
        for (int j = 0; j < plainTextData.Length; j++)
        {
            cipherTextData[j] = cipher.ReturnByte(plainTextData[j]);
        }
       
        return cipherTextData;        
    }
    

    /// <summary>
    /// Example 13 – ChaCha Stream Cipher  Deccryption
    /// </summary>        
    public static byte[] chachaStreamDecrypt(ICipherParameters keyParamWithIV, byte[] cipherTextData)
    {
        IStreamCipher cipher = new ChaChaEngine();        
        cipher.Init(false, keyParamWithIV);
        byte[] plainTextData = new byte[cipherTextData.Length];

        // simulate stream
        for (int j = 0; j < cipherTextData.Length; j++)
        {
            plainTextData[j] = cipher.ReturnByte(cipherTextData[j]);
        }
        return plainTextData;
    }

    /// <summary>
    /// Example 14 – HC128 Stream Cipher Encryption
    /// </summary>    
    public static byte[] h128StreamEncrypt(ICipherParameters keyParamWithIV, byte[] plainTextData)
    {
        IStreamCipher cipher = new HC128Engine();
        cipher.Init(true, keyParamWithIV);

        byte[] cipherTextData = new byte[plainTextData.Length];
        
        // simulate stream
        for (int j = 0; j < plainTextData.Length; j++)
        {
            cipherTextData[j] = cipher.ReturnByte(plainTextData[j]);
        }
       
        return cipherTextData;        
    }
    
    /// <summary>
    /// Example 14 – HC128 Stream Cipher Decryption
    /// </summary>    
    public static byte[] h128StreamDecrypt(ICipherParameters keyParamWithIV, byte[] cipherTextData)
    {
        IStreamCipher cipher = new HC128Engine();
        cipher.Init(true, keyParamWithIV);

        byte[] plainTextData = new byte[cipherTextData.Length];
        
        // simulate stream
        for (int j = 0; j < cipherTextData.Length; j++)
        {
            plainTextData[j] = cipher.ReturnByte(cipherTextData[j]);
        }
       
        return plainTextData;        
    }
    
}
